# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.interpolate as spinterp
import scipy.integrate as spinteg
from scipy.optimize import leastsq
import os
from .. import Points, ScalarField, VectorField, Profile, \
    TemporalVectorFields, TemporalScalarFields
from ..utils.types import ARRAYTYPES, NUMBERTYPES, STRINGTYPES
from ..utils import RemoveFortranOutput

### Gradients based operation
def get_gradients(field, raw=False):
    """
    Return gradients along x and y.

    (Obtained arrays corespond to components of the Jacobian matrix)

    Parameters
    ----------
    vf : VelocityField, ScalarField or Profile object
        Field/profile to compute gradient from.
    raw : boolean
        If 'False' (default), ScalarFields objects are returned.
        If 'True', arrays are returned.

    Returns
    -------
    grad : tuple of ScalarField or arrays or Profile
        For VectorField input : (dVx/dx, dVx/dy, dVy/dx, dVy/dy),
        for ScalarField input : (dV/dx, dV/dy).
        for Profile input : dy/dx.
    """
    if isinstance(field, ScalarField):
        dx = field.axe_x[1] - field.axe_x[0]
        dy = field.axe_y[1] - field.axe_y[0]
        grad_x, grad_y = np.gradient(np.ma.masked_array(field.values,
                                                        field.mask),
                                     dx, dy)
        # applying masks
        mask = np.logical_or(grad_x.mask, grad_y.mask)
        if raw:
            # returning
            grad_x.mask = mask
            grad_y.mask = mask
            return grad_x, grad_y
        else:
            # arranging units
            gradx = ScalarField()
            unit_values_x = field.unit_values/field.unit_x
            factx = unit_values_x.asNumber()
            unit_values_x = unit_values_x/factx
            grad_x *= factx
            unit_values_y = field.unit_values/field.unit_y
            facty = unit_values_y.asNumber()
            unit_values_y = unit_values_y/facty
            grad_y *= facty
            # returning
            gradx.import_from_arrays(field.axe_x, field.axe_y, grad_x.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_x)
            grady = ScalarField()
            grady.import_from_arrays(field.axe_x, field.axe_y, grad_y.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_y)
            return gradx, grady
    elif isinstance(field, VectorField):
        dx = field.axe_x[1] - field.axe_x[0]
        dy = field.axe_y[1] - field.axe_y[0]
        Vx_dx, Vx_dy = np.gradient(np.ma.masked_array(field.comp_x,
                                                      field.mask),
                                   dx, dy)
        Vy_dx, Vy_dy = np.gradient(np.ma.masked_array(field.comp_y,
                                                      field.mask),
                                   dx, dy)
        # applying masks
        mask = np.logical_or(Vx_dx.mask, Vx_dy.mask)
        mask = np.logical_or(mask, Vy_dx.mask)
        mask = np.logical_or(mask, Vy_dy.mask)
        if raw:
            # returning
            Vx_dx.mask = mask
            Vx_dy.mask = mask
            Vy_dx.mask = mask
            Vy_dy.mask = mask
            return (Vx_dx, Vx_dy, Vy_dx, Vy_dy)
        else:
            #arragning unit
            unit_values_x = field.unit_values/field.unit_x
            factx = unit_values_x.asNumber()
            unit_values_x /= factx
            Vx_dx *= factx
            Vy_dx *= factx
            unit_values_y = field.unit_values/field.unit_y
            facty = unit_values_y.asNumber()
            unit_values_y /= facty
            Vx_dy *= facty
            Vy_dy *= facty
            #returning
            grad1 = ScalarField()
            grad1.import_from_arrays(field.axe_x, field.axe_y, Vx_dx.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_x)
            grad2 = ScalarField()
            grad2.import_from_arrays(field.axe_x, field.axe_y, Vx_dy.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_y)
            grad3 = ScalarField()
            grad3.import_from_arrays(field.axe_x, field.axe_y, Vy_dx.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_x)
            grad4 = ScalarField()
            grad4.import_from_arrays(field.axe_x, field.axe_y, Vy_dy.data,
                                     mask=mask, unit_x=field.unit_x,
                                     unit_y=field.unit_y,
                                     unit_values=unit_values_y)
            return grad1, grad2, grad3, grad4
    elif isinstance(field, Profile):
        tmp_prof = field.evenly_space('quadratic')
        axe_x = tmp_prof.x
        dx = axe_x[1] - axe_x[0]
        mask = tmp_prof.mask
        grad = np.gradient(tmp_prof.y)
        if raw:
            return grad
        else:
            unit_y = tmp_prof.unit_y/tmp_prof.unit_x
            facty = unit_y.asNumber()
            unit_y /= facty
            grad *= facty
            prof_grad = Profile(axe_x, grad, mask=mask,
                                unit_x=tmp_prof.unit_x, unit_y=unit_y)
            return prof_grad
    else:
        raise TypeError()


def reconstruct_from_gradients(field_dx, field_dy, field2_dx=None,
                               field2_dy=None, ols=False, maxiter=10):
    """
    Reconstruct a field with the gradients of this field.

    Parameters
    ----------
    field_dx, field_dy : ScalarField objects
        Gradients along x and y.
    field2_dx, field_dy : ScalarField objects, optional
        Gradients for the second component, if the reconstructed field is a
        vector field.
    ols : boolean, optional
        If 'True', ordinary least square is used to get a more precise result
        (can be quite long).
        If 'False' (default), a simple reconstruction based on taylor
        developement is used.
    maxiter : integer, optional
        Maximum number of iteration for the ols solver (default: 10)
        (more mean accurate results but slower computation).


    Returns
    -------
    rec_field : ScalarField or VectorField object
        Reconstructed field.

    Notes
    -----
    Given result can only be relative values
    (because of the information lost while derivating).
    The returned result are normalized so that the mean of the field is egal
    to zero.
    """
    ### checking parameters
    if not isinstance(field_dx, ScalarField):
        raise TypeError()
    if not isinstance(field_dy, ScalarField):
        raise TypeError()
    if field2_dx is None and field2_dy is None:
        kind = 'SF'
        if field2_dx is not None or field2_dy is not None:
            raise ValueError()
    else:
        kind = 'VF'
        if not isinstance(field2_dx, ScalarField):
            raise TypeError()
        if not isinstance(field2_dy, ScalarField):
            raise TypeError()
    ### recursive loop for VF
    if kind == 'VF':
        Vx = reconstruct_from_gradients(field_dx, field_dy, ols=ols)
        Vy = reconstruct_from_gradients(field2_dx, field2_dy, ols=ols)
        mask = np.logical_or(Vx.mask, Vy.mask)
        V = VectorField()
        V.import_from_arrays(Vx.axe_x, Vx.axe_y, Vx.values, Vy.values,
                             mask=mask, unit_x=Vx.unit_x, unit_y=Vx.unit_y,
                             unit_values=Vx.unit_values)
        return V
    ### reconstruction for SF
    else:
        ### getting data
        axe_x = field_dx.axe_x
        axe_y = field_dy.axe_y
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        du_dx = field_dx.values
        du_dy = field_dy.values
        unit_values = field_dx.unit_values*field_dx.unit_x
        fact_unit_values = unit_values.asNumber()
        unit_values /= unit_values.asNumber()

        ###  Getting borders
        # 1 (0, 0)
        u1 = np.zeros(du_dx.shape)
        for i in np.arange(1, u1.shape[0]):
            u1[i, 0] = u1[i - 1, 0] + dx*(du_dx[i - 1, 0]
                                          + du_dx[i, 0])/2.
        for i in np.arange(1, u1.shape[1]):
            u1[0, i] = u1[0, i - 1] + dy*(du_dy[0, i - 1]
                                          + du_dy[0, i])/2.
        # 2(0, 1)
        u2 = np.zeros(du_dx.shape)
        for i in np.arange(1, u2.shape[0]):
            u2[i, -1] = u2[i - 1, -1] + dx*(du_dx[i - 1, -1]
                                            + du_dx[i, -1])/2.
        for i in np.arange(u2.shape[1] - 2, -1, -1):
            u2[0, i] = u2[0, i + 1] - dy*(du_dy[0, i + 1]
                                          + du_dy[0, i])/2.
        # 3 (1, 0)
        u3 = np.zeros(du_dx.shape)
        for i in np.arange(u3.shape[0] - 2, -1, -1):
            u3[i, 0] = u3[i + 1, 0] - dx*(du_dx[i + 1, 0]
                                          + du_dx[i, 0])/2.
        for i in np.arange(1, u3.shape[1]):
            u3[-1, i] = u3[-1, i - 1] + dy*(du_dy[-1, i - 1]
                                            + du_dy[-1, i])/2.
        # 4 (1, 1)
        u4 = np.zeros(du_dx.shape)
        for i in np.arange(u4.shape[0] - 2, -1, -1):
            u4[i, -1] = u4[i + 1, -1] - dx*(du_dx[i + 1, -1]
                                            + du_dx[i, -1])/2.
        for i in np.arange(u4.shape[1] - 2, -1, -1):
            u4[-1, i] = u4[-1, i + 1] - dy*(du_dy[-1, i + 1]
                                            + du_dy[-1, i])/2.
        # concatenate borders
        filt_border = np.zeros(u1.shape, dtype=bool)
        filt_border[:, 0] = True
        filt_border[:, -1] = True
        filt_border[0, :] = True
        filt_border[-1, :] = True
        u4 = (u4 - (u4[-1, 0] - u1[-1, 0] + u4[0, -1] - u1[0, -1])/2.)
        u4[0:-1, 0:-1] = 0.
        ua = u1 + u4
        ua[-1, 0] /= 2.
        ua[0, -1] /= 2.
        ua[filt_border] -= np.mean(ua[filt_border])
        u3 = (u3 - (u3[0, 0] - u2[0, 0] + u3[-1, -1] - u2[-1, -1])/2.)
        u3[0:-1, 1::] = 0.
        ub = u2 + u3
        ub[0, 0] /= 2.
        ub[-1, -1] /= 2.
        ub[filt_border] -= np.mean(ub[filt_border])
        u = (ua + ub)/2.
        ### Getting center
        # 1 from (0, 0)
        u1 = u.copy()
        for i in np.arange(1, u1.shape[0]):
            for j in np.arange(1, u1.shape[1]):
                est_a = u1[i - 1, j] + dx*(du_dx[i - 1, j]
                                           + du_dx[i, j])/2.
                est_b = u1[i, j - 1] + dy*(du_dy[i, j - 1]
                                           + du_dy[i, j])/2.
                u1[i, j] = (est_a + est_b)/2.
        # 2 from (0, 1)
        u2 = u.copy()
        for i in np.arange(1, u2.shape[0]):
            for j in np.arange(u2.shape[1] - 2, -1, -1):
                est_a = u2[i - 1, j] + dx*(du_dx[i - 1, j]
                                           + du_dx[i, j])/2.
                est_b = u2[i, j + 1] - dy*(du_dy[i, j + 1]
                                           + du_dy[i, j])/2.
                u2[i, j] = (est_a + est_b)/2.
        # 3 from (1, 0)
        u3 = u.copy()
        for i in np.arange(u3.shape[0] - 2, -1, -1):
            for j in np.arange(1, u3.shape[1]):
                est_a = u3[i + 1, j] - dx*(du_dx[i + 1, j]
                                           + du_dx[i, j])/2.
                est_b = u3[i, j - 1] + dy*(du_dy[i, j - 1]
                                           + du_dy[i, j])/2.
                u3[i, j] = (est_a + est_b)/2.
        # 4 from (1, 1)
        u4 = u.copy()
        for i in np.arange(u4.shape[0] - 2, -1, -1):
            for j in np.arange(u2.shape[1] - 2, -1, -1):
                est_a = u4[i + 1, j] - dx*(du_dx[i + 1, j]
                                           + du_dx[i, j])/2.
                est_b = u4[i, j + 1] - dy*(du_dy[i, j + 1]
                                           + du_dy[i, j])/2.
                u4[i, j] = (est_a + est_b)/2.
        # normalize
        filt_corn = np.zeros(u.shape, dtype=int)
        filt_corn[0, 0] = 1
        filt_corn[0, -1] = 1
        filt_corn[-1, 0] = 1
        filt_corn[-1, -1] = 1
        u1 -= np.mean(u1[filt_corn])
        u2 -= np.mean(u2[filt_corn])
        u3 -= np.mean(u3[filt_corn])
        u4 -= np.mean(u4[filt_corn])
        # concatenate (with linear weight fonction)
        ### TODO : let or remove ?
        X, Y = np.meshgrid(np.arange(u.shape[1]), np.arange(u.shape[0]))
        weight1 = X[::-1, ::-1] + Y[::-1, ::-1]
        weight2 = X[::-1, :] + Y[::-1, :]
        weight3 = X[:, ::-1] + Y[:, ::-1]
        weight4 = X + Y
        weight_tot = weight1 + weight2 + weight3 + weight4
        u = (u1*weight1 + u2*weight2 + u3*weight3 + u4*weight4)/weight_tot

        ### Second OLS reconstruction
        if ols:
            def min_funct(u, du_dx, du_dy, dx, dy):
                u = u.reshape(du_dx.shape)
                du2_dx, du2_dy = np.gradient(u, dx, dy)
                res = np.abs(du_dx - du2_dx) + np.abs(du_dy - du2_dy)
                return res.flatten()
            u, mess = leastsq(min_funct, u.flatten(),
                              args=(du_dx, du_dy, dx, dy),
                              maxfev=maxiter*len(u.flatten()))
            u = u.reshape(du_dx.shape)
        ### retuning
        u *= fact_unit_values
        u = u - np.mean(u)
        U = ScalarField()
        U.import_from_arrays(axe_x, axe_y, u,
                             unit_x=field_dx.unit_x, unit_y=field_dx.unit_y,
                             unit_values=unit_values)
#        if ols:
#            U.smooth(tos='gaussian', size=1, inplace=True)
        return U


def get_jacobian_eigenproperties(field, raw=False, eig_val=True,
                                 eig_vect=True):
    """
    Return eigenvalues and eigenvectors of the jacobian matrix on all the
    field.

    Parameters
    ----------
    field : VectorField object
        .
    raw : boolean
        If 'False' (default), ScalarFields objects are returned.
        If 'True', arrays are returned.
    eig_val : boolean, optional
        If 'True', eigenvalues are returned.
    eig_vect : boolean, optional
        If 'True', eigenvectors are returned.

    Returns
    -------
    eig_val1_re : ScalarField object, or array
        Real part of the first eigenvalue.
    eig_val1_im : ScalarField object, or array
        Imaginary part of the first eigenvalue.
    eig_val2_re : ScalarField object, or array
        Real part of the second eigenvalue.
    eig_val2_im : ScalarField object, or array
        Imaginary part of the second eigenvalue.
    eig1_vf : VectorField object, or tuple of arrays
        Eigenvector associated with first eigenvalue.
    eig2_vf : VectorField object, or tuple of arrays
        Eigenvector associated with second eigenvalue.
    """
    # getting datas
    Vx_dx, Vx_dy, Vy_dx, Vy_dy = get_gradients(field, raw=True)
    shape = Vx_dx.shape
    mask = Vx_dx.mask
    mask = np.logical_or(mask, Vx_dy.mask)
    mask = np.logical_or(mask, Vy_dx.mask)
    mask = np.logical_or(mask, Vy_dy.mask)
    Vx_dx = Vx_dx.data
    Vx_dy = Vx_dy.data
    Vy_dx = Vy_dx.data
    Vy_dy = Vy_dy.data
    # loop on flatten arrays
    eig1 = np.zeros(shape, dtype=complex)
    eig2 = np.zeros(shape, dtype=complex)
    eig1v_x = np.zeros(shape)
    eig1v_y = np.zeros(shape)
    eig2v_x = np.zeros(shape)
    eig2v_y = np.zeros(shape)
    for i in np.arange(shape[0]*shape[1]):
        # breaking when masked
        if mask.flat[i]:
            continue
        # getting the local jacobian matrix
        loc_jac = [[Vx_dx.flat[i], Vx_dy.flat[i]],
                   [Vy_dx.flat[i], Vy_dy.flat[i]]]
        # getting local max eigenvalue and associated eigenvectors
        loc_eigv, loc_eigvect = np.linalg.eig(loc_jac)
        max_eig = np.argmax(loc_eigv)
        if max_eig == 0:
            min_eig = 1
        else:
            min_eig = 0
        # storing in arrays
        eig1.flat[i] = loc_eigv[max_eig]
        eig2.flat[i] = loc_eigv[min_eig]
        eig1v_x.flat[i] = loc_eigvect[0, max_eig]
        eig1v_y.flat[i] = loc_eigvect[1, max_eig]
        eig2v_x.flat[i] = loc_eigvect[0, min_eig]
        eig2v_y.flat[i] = loc_eigvect[1, min_eig]
    # extracting real and imaginary
    eig1_r = np.real(eig1)
    eig1_i = np.imag(eig1)
    eig2_r = np.real(eig2)
    eig2_i = np.imag(eig2)
    # returning
    if raw:
        ret = ()
        if eig_val:
            ret += (eig1_r, eig1_i, eig2_r, eig2_i)
        if eig_vect:
            ret += ((eig1v_x, eig1v_y), (eig2v_x, eig2v_y))
        return ret
    else:
        ret = ()
        if eig_val:
            eig1_re_sf = ScalarField()
            eig1_re_sf.import_from_arrays(field.axe_x, field.axe_y, eig1_r,
                                          mask=mask, unit_x=field.unit_x,
                                          unit_y=field.unit_y,
                                          unit_values="")
            eig1_im_sf = ScalarField()
            eig1_im_sf.import_from_arrays(field.axe_x, field.axe_y, eig1_i,
                                          mask=mask, unit_x=field.unit_x,
                                          unit_y=field.unit_y,
                                          unit_values="")
            eig2_re_sf = ScalarField()
            eig2_re_sf.import_from_arrays(field.axe_x, field.axe_y, eig2_r,
                                          mask=mask, unit_x=field.unit_x,
                                          unit_y=field.unit_y,
                                          unit_values="")
            eig2_im_sf = ScalarField()
            eig2_im_sf.import_from_arrays(field.axe_x, field.axe_y, eig2_i,
                                          mask=mask, unit_x=field.unit_x,
                                          unit_y=field.unit_y,
                                          unit_values="")
            ret += (eig1_re_sf, eig1_im_sf, eig2_re_sf, eig2_im_sf)
        if eig_vect:
            eig1_vf = VectorField()
            eig1_vf.import_from_arrays(field.axe_x, field.axe_y, eig1v_x, eig1v_y,
                                       mask=mask, unit_x=field.unit_x,
                                       unit_y=field.unit_y,
                                       unit_values="")
            eig2_vf = VectorField()
            eig2_vf.import_from_arrays(field.axe_x, field.axe_y, eig2v_x, eig2v_y,
                                       mask=mask, unit_x=field.unit_x,
                                       unit_y=field.unit_y,
                                       unit_values="")
            ret += (eig1_vf, eig2_vf)
        return ret


def get_Kenwright_field(field, raw=False):
    """
    Return a field with the vector product between the velocity field and the
    eigen vectors of the velocity jacobian.
    Values of 0 on these field should represent separation lines
    or re-attachment lines
    (See Kenwright et al (1998)).

    Parameters
    ----------
    field : VectorField
        .
    raw : bool, optional
        If 'False' (default), ScalarFields are returned
        If 'True', arrays are returned

    Returns
    -------
    K1_field : ScalarField
        Vector product with the principal eigen vector
    K2_field : ScalarField
        Vector product with the oher eigen vector
    """
    # get jacobian  eigen properties
    eigval1, eigval1_im, eigval2, eigval2_im, eigvect1, eigvect2 \
        = get_jacobian_eigenproperties(field)
    K1 = np.zeros(field.shape)
    K2 = np.zeros(field.shape)
    # making vector product
    for xi in np.arange(len(field.axe_x)):
        for yi in np.arange(len(field.axe_y)):
            K1[xi, yi] = (eigvect1.comp_x[xi, yi]*field.comp_y[xi, yi]
                          - eigvect1.comp_y[xi, yi]*field.comp_x[xi, yi])
            K2[xi, yi] = (eigvect2.comp_x[xi, yi]*field.comp_y[xi, yi]
                          - eigvect2.comp_y[xi, yi]*field.comp_x[xi, yi])
    # returning
    if raw:
        return K1, K2
    else:
        K1_sf = ScalarField()
        K1_sf.import_from_arrays(field.axe_x, field.axe_y, K1,
                                 mask=np.isnan(K1),
                                 unit_x=field.unit_x, unit_y=field.unit_y,
                                 unit_values=field.unit_values)
        K2_sf = ScalarField()
        K2_sf.import_from_arrays(field.axe_x, field.axe_y, K2,
                                 mask=np.isnan(K2),
                                 unit_x=field.unit_x, unit_y=field.unit_y,
                                 unit_values=field.unit_values)
        return K1_sf, K2_sf


def get_grad_field(field, direction='x'):
    """
    Return a field based on original field gradients.
    (Vx = dV/dx, Vy = DV/Vy)

    Parameters
    ----------
    field : VectorField object
        .
    direction : string in ['x', 'y']
        if 'x', return Vx gradients
        if 'y', return Vy gradients.

    Returns
    -------
    gfield : VectorField object
        .
    """
    # checking parameters:
    if not direction in ['x', 'y']:
        raise ValueError()
    # getting gradients
    field.comp_x = field.comp_x
    field.comp_y = field.comp_y
    grads = get_gradients(field)
    # returning
    if direction == 'x':
        gvx = grads[0]
        gvy = grads[1]
    else:
        gvx = grads[2]
        gvy = grads[3]
    gfield = VectorField()
    gfield.import_from_arrays(field.axe_x, field.axe_y, comp_x=gvx.values,
                              comp_y=gvy.values, unit_x=field.unit_x,
                              unit_y=field.unit_y,
                              unit_values=field.unit_values)
    return gfield


def get_track_field(vf):
    """
    Return the track field.
    (Vx, Vy) => (-Vy, Vx)
    """
    Vx = -vf.comp_y
    Vy = vf.comp_x
    track_field = VectorField()
    track_field.import_from_arrays(vf.axe_x, vf.axe_y, Vx, Vy, mask=vf.mask,
                                   unit_x=vf.unit_x, unit_y=vf.unit_y,
                                   unit_values=vf.unit_values)
    return track_field

def get_divergence(vf):
    """
    Return the divergence.
    """
    if isinstance(vf, VectorField):
        dudx, dudy, dvdx, dvdy = get_gradients(vf)
        return dudx + dvdy
    elif isinstance(vf, TemporalVectorFields):
        tmp_vf = TemporalScalarFields()
        for i, _ in enumerate(vf.fields):
            tmp_vf.add_field(get_divergence(vf.fields[i]), time=vf.times[i],
                             unit_times=vf.unit_times)
        return tmp_vf
    else:
        raise TypeError()


### Stream and track lines
def get_streamlines_fast(vf, xy, delta=.25, interp='linear',
                         reverse=False):
    """
    Return a tuples of Points object representing the streamline begining
    at the points specified in xy.
    Is called fast because use simpler integration method than the
    'get_streamlines()' method and can be tuned with the 'delta' and 'interp'
    parameters.

    Parameters
    ----------
    vf : VectorField or velocityField object
        Field on which compute the streamlines
    xy : tuple
        Tuple containing each starting point for streamline.
    delta : number, optional
        Spatial discretization of the stream lines,
        relative to a the spatial discretization of the field.
    interp : string, optional
        Used interpolation for streamline computation.
        Can be 'linear'(default) or 'cubic'
    reverse : boolean, optional
        If True, the streamline goes upstream.

    Returns
    -------
    streams : tuple of Points object
        Each Points object represent a streamline

    """
    # checking parameters coherence
    if not isinstance(vf, VectorField):
        raise TypeError()
    axe_x, axe_y = vf.axe_x, vf.axe_y
    if not isinstance(xy, ARRAYTYPES):
        raise TypeError("'xy' must be a tuple of arrays")
    xy = np.array(xy, dtype=float)
    if xy.shape == (2,):
        xy = [xy]
    elif len(xy.shape) == 2 and xy.shape[1] == 2:
        pass
    else:
        raise ValueError("'xy' must be a tuple of arrays")
    if not isinstance(delta, NUMBERTYPES):
        raise TypeError("'delta' must be a number")
    if delta <= 0:
        raise ValueError("'delta' must be positive")
    if delta > len(axe_x) or delta > len(axe_y):
        raise ValueError("'delta' is too big !")
    if not isinstance(interp, STRINGTYPES):
        raise TypeError("'interp' must be a string")
    if not (interp == 'linear' or interp == 'cubic'):
        raise ValueError("Unknown interpolation method")
    # getting datas
    tmpvf = vf.copy()
    tmpvf.fill()
    unit_x, unit_y = vf.unit_x, vf.unit_y
    Vx, Vy = vf.comp_x, vf.comp_y
    deltaabs = delta * ((axe_x[-1]-axe_x[0])/len(axe_x)
                        + (axe_y[-1]-axe_y[0])/len(axe_y))/2.
    deltaabs2 = deltaabs**2
    # interpolation du champ de vitesse
    if interp == 'linear':
        interp_vx = spinterp.RectBivariateSpline(axe_x, axe_y, Vx,
                                                 kx=1, ky=1)
        interp_vy = spinterp.RectBivariateSpline(axe_x, axe_y, Vy,
                                                 kx=1, ky=1)

    elif interp == 'cubic':
        interp_vx = spinterp.RectBivariateSpline(axe_x, axe_y, Vx,
                                                 kx=3, ky=3)
        interp_vy = spinterp.RectBivariateSpline(axe_x, axe_y, Vy,
                                                 kx=3, ky=3)
    # Calcul des streamlines
    streams = []
    longmax = int(np.round((len(axe_x) + len(axe_y))/delta))
    for coord in xy:
        # getting coordinates
        x = coord[0]
        y = coord[1]
        x = float(x)
        y = float(y)
        # si le points est en dehors, on ne fait rien
        if x < axe_x[0] or x > axe_x[-1] or y < axe_y[0] or y > axe_y[-1]:
            continue
        # initialisation du calcul d'une streamline
        stream = np.zeros((longmax, 2))
        stream[0, :] = [x, y]
        i = 1
        # calcul d'une streamline
        while True:
            tmp_vx = interp_vx(stream[i-1, 0], stream[i-1, 1])[0, 0]
            tmp_vy = interp_vy(stream[i-1, 0], stream[i-1, 1])[0, 0]
            norm = np.linalg.norm([tmp_vx, tmp_vy])
            # tests d'arret
            if tmp_vx == 0 and tmp_vy == 0:
                # if masked value
                break
            if i >= longmax-1:
                break
            if i > 15:
                no = [stream[0:i-10, 0] - stream[i-1, 0],
                      stream[0:i-10, 1] - stream[i-1, 1]]
                no = no[0]**2 + no[1]**2
                if any(no < deltaabs2/2):
                    break
            # calcul des dx et dy pour une streamline
            if reverse:
                norm = -norm
            dx = tmp_vx/norm*deltaabs
            dy = tmp_vy/norm*deltaabs
            stream[i, :] = [stream[i-1, 0] + dx, stream[i-1, 1] + dy]
            i += 1
            # tests d'arret
            x = stream[i-1, 0]
            y = stream[i-1, 1]
            if (x < axe_x[0] or x > axe_x[-1] or y < axe_y[0]
                    or y > axe_y[-1]):
                break
        # creating Points objects for returning
        stream = stream[:i]
        pts = Points(stream, unit_x=unit_x, unit_y=unit_y,
                     name='streamline at x={:.3f}, y={:.3f}'.format(x, y))
        streams.append(pts)
    # returning
    if len(streams) == 0:
        return None
    elif len(streams) == 1:
        return streams
    else:
        return streams

def get_streamlines(VF, xys, reverse=False, rel_err=1.e-8,
                    max_steps=1000, resolution=0.25, resolution_kind='length',
                    boundary_tr='stop'):
    """
    Return the lagrangien displacement of a set of particules initialy at the
    positions xys.
    Use Fortran integration of the VODE algorithm.
    (Source: http://www.netlib.org/ode/vode.f)

    Parameters
    ----------
    VF : VectorField or velocityField object
        Field on which compute the streamlines
    xys : 2xN tuple of number
        Initial points positions
    rel_err : number
        relative maximum error for rk4 algorithm
    max_steps : integer
        Maximum number of steps (default = 1000).
    resolution : number,
        resolution of the resulting streamline (do not impact accuracy).
    resolution_kind : string in ['time', 'length']
        if 'time', returned points time space are homogeneous,
        if 'length', returned points space interval are homogeneous.
    boundary_tr : string
        Method to treat the field boundaries. If 'stop', streamlines are
        stopped when encountering a boundary, If 'hide', streamlines
        that encounter a boundary are not returned.

    Returns
    -------
    streamlines: tuple of Points objects
        Wanted streamlines
    """

    #############
    ### check ###
    #############
    if not isinstance(VF, VectorField):
        raise TypeError()
    if np.any(VF.mask):
        masked_values = True
    else:
        masked_values = False
    if not isinstance(xys, ARRAYTYPES):
        raise TypeError()
    xys = np.array(xys)
    if xys.shape == (2,):
        xys = np.array([xys], subok=True, dtype=float)
    if xys.ndim != 2:
        raise ValueError
    if not isinstance(reverse, (bool, np.bool, np.bool_)):
        raise TypeError()
    if not isinstance(rel_err, NUMBERTYPES):
        raise TypeError()
    if rel_err <= 0:
        raise ValueError()
    if not isinstance(max_steps, NUMBERTYPES):
        raise TypeError()
    max_steps = int(max_steps)
    if not isinstance(boundary_tr, STRINGTYPES):
        raise TypeError()
    if boundary_tr not in ['stop', 'hide']:
        raise ValueError()

    ################
    ### get data ###
    ################
    from scipy.interpolate import RectBivariateSpline
    axe_x, axe_y = VF.axe_x, VF.axe_y
    # AD
    fact_axe = (axe_x[-1] - axe_x[0])/(axe_y[-1] - axe_y[0])
    xys[:, 0] = (xys[:, 0] - axe_x[0])/(axe_x[-1] - axe_x[0])
    xys[:, 1] = (xys[:, 1] - axe_y[0])/(axe_y[-1] - axe_y[0])
    axe_x = (axe_x - axe_x[0])/(axe_x[-1] - axe_x[0])
    axe_y = (axe_y - axe_y[0])/(axe_y[-1] - axe_y[0])
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    dxy = np.mean([dx, dy])
    ad_vel = np.max(np.abs(VF.magnitude[~VF.mask]))
    vx = VF.comp_x/ad_vel/fact_axe
    vy = VF.comp_y/ad_vel
    mask = VF.mask
    vx[mask] = 0
    vy[mask] = 0
    # get interpolators
    Vx_tor = RectBivariateSpline(axe_x, axe_y, vx,
                                 kx=1, ky=1)
    Vy_tor = RectBivariateSpline(axe_x, axe_y, vy,
                                 kx=1, ky=1)
    # velocity function
    if reverse:
        def fun(xy):
            if is_outside(axe_x, axe_y, xy):
                raise ValueError()
            if masked_values:
                ind_x = int(round(xy[0]/dx))
                ind_y = int(round(xy[1]/dy))
                if mask[ind_x, ind_y]:
                    raise ValueError()
            return -np.array([Vx_tor(*xy)[0][0],
                              Vy_tor(*xy)[0][0]])
    else:
        def fun(xy):
            if is_outside(axe_x, axe_y, xy):
                raise ValueError()
            if masked_values:
                ind_x = int(round(xy[0]/dx))
                ind_y = int(round(xy[1]/dy))
                if mask[ind_x, ind_y]:
                    raise ValueError()
            return np.array([Vx_tor(*xy)[0][0],
                             Vy_tor(*xy)[0][0]])
    if reverse:
        def fun2(t, xy):
            return -np.array([Vx_tor(*xy)[0][0],
                              Vy_tor(*xy)[0][0]])
    else:
        def fun2(t, xy):
            return np.array([Vx_tor(*xy)[0][0],
                             Vy_tor(*xy)[0][0]])

    ####################
    ### outside test ###
    ####################
    def is_outside(axe_x, axe_y, pt):
        if pt[0] < axe_x[0]:
            outside = True
        elif pt[0] > axe_x[-1]:
            outside = True
        elif pt[1] < axe_y[0]:
            outside = True
        elif pt[1] > axe_y[-1]:
            outside = True
        else:
            outside = False
        return outside


    ########################
    ### Loop on points   ###
    ########################
    streams = []
    for xy in xys:
        # check for invalid points
        if is_outside(axe_x, axe_y, xy):
            continue

        ############################
        ### solve streamline ODE ###
        ############################
        new_xys = [xy]
        new_ts = [0]
        with RemoveFortranOutput():
            ODE = spinteg.ode(fun2)
            ODE.set_integrator('vode')
            ODE.set_initial_value(xy)
            res = [xy]
            nmb_it = 0
            while ODE.successful():
                try:
                    if resolution_kind == "time":
                        rk_dt = resolution
                    else:
                        rk_dt = dxy/np.linalg.norm(fun(res[-1]))*resolution
                except (ValueError, RuntimeWarning):
                    break
                try:
                    tmp_xy = ODE.integrate(ODE.t + rk_dt)
                except UserWarning:
                    break
                nmb_it += 1
                # stopping tests (max iteration)
                if nmb_it > max_steps:
                    break
                if resolution_kind == "time":
                    min_dxy = (dxy/40)**2
                    if (tmp_xy[0] - res[-1][0])**2 + (tmp_xy[1] - res[-1][1])**2 < min_dxy:
                        break
                # store
                new_ts.append(new_ts[-1] + rk_dt)
                res.append(tmp_xy)
            new_xys = res
#        warnings.filterwarnings('once')
        # storing and rescaling
        if len(new_xys) != 0:
            new_xys = np.array(new_xys, dtype=float)
            new_xys[:, 0] = VF.axe_x[0] + new_xys[:, 0]*(VF.axe_x[-1] - VF.axe_x[0])
            new_xys[:, 1] = VF.axe_y[0] + new_xys[:, 1]*(VF.axe_y[-1] - VF.axe_y[0])
        stream = Points(new_xys, unit_x=VF.unit_x, unit_y=VF.unit_y,
                        v=new_ts)
        streams.append(stream)

    #################
    ### returning ###
    #################
    if len(streams) == 1:
        return streams[0]
    else:
        return streams

def get_fieldlines(VF, xys, reverse=False, rel_err=1.e-8,
                   max_steps=1000, resolution=0.25, resolution_kind='length',
                   boundary_tr='stop'):
    """
    Return the field lines associated to a set of particules initialy at the
    positions xys.
    Use Fortran integration of the VODE algorithm.
    (Source: http://www.netlib.org/ode/vode.f)

    Parameters
    ----------
    VF : VectorField or velocityField object
        Field on which compute the fieldlines
    xys : 2xN tuple of number
        Initial points positions
    rel_err : number
        relative maximum error for rk4 algorithm
    max_steps : integer
        Maximum number of steps (default = 1000).
    resolution : number,
        resolution of the resulting fieldline (do not impact accuracy).
    resolution_kind : string in ['time', 'length']
        if 'time', returned points time space are homogeneous,
        if 'length', returned points space interval are homogeneous.
    boundary_tr : string
        Method to treat the field boundaries. If 'stop', fieldlines are
        stopped when encountering a boundary, If 'hide', fieldlines
        that encounter a boundary are not returned.
    """
    # transform the field so that streamlines become filedlines
    tmp_vf = VF.copy()
    tmp_vf.comp_x = VF.comp_y
    tmp_vf.comp_y = -VF.comp_x
    # get the fieldlines
    sts = get_streamlines(tmp_vf, xys, reverse=reverse, rel_err=rel_err, max_steps=max_steps,
                          resolution=resolution, resolution_kind=resolution_kind,
                          boundary_tr=boundary_tr)
    # return
    return sts


def get_tracklines(VF, xys, reverse=False, rel_err=1.e-8,
                    max_steps=1000, resolution=0.25, resolution_kind='length',
                    boundary_tr='stop'):
    """
    Return a tuples of Points object representing the tracklines begining
    at the points specified in xy.
    Warning : fill the field before computing streamlines, can give bad
    results if the field have a lot of masked values.

    Parameters
    ----------
    VF : VectorField or velocityField object
        Field on which compute the streamlines
    xys : 2xN tuple of number
        Initial points positions
    rel_err : number
        relative maximum error for rk4 algorithm
    max_steps : integer
        Maximum number of steps (default = 1000).
    resolution : number,
        resolution of the resulting streamline (do not impact accuracy).
    resolution_kind : string in ['time', 'length']
        if 'time', returned points time space are homogeneous,
        if 'length', returned points space interval are homogeneous.
    boundary_tr : string
        Method to treat the field boundaries. If 'stop', streamlines are
        stopped when encountering a boundary, If 'hide', streamlines
        that encounter a boundary are not returned.

    Returns
    -------
    tracks : tuple of Points object
        Each Points object represent a trackline

    """
    # get the track field
    track_field = get_track_field(VF)
    return get_streamlines(track_field, xys, rel_err=rel_err,
                           max_steps=max_steps, resolution=resolution,
                           resolution_kind=resolution_kind,
                           boundary_tr=boundary_tr)

def get_shear_stress(vf, raw=False):
    """
    Return a vector field with the shear stress.

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return two arrays,
        if 'False' (default), return a VectorField object.
    """
    if not isinstance(vf, VectorField):
        raise TypeError()
    tmp_vf = vf.copy()
    tmp_vf.crop_masked_border(inplace=True)
    tmp_vf.fill()
    # Getting gradients and axes
    axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
    comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
    mask = tmp_vf.mask
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    du_dx, du_dy = np.gradient(comp_x, dx, dy)
    dv_dx, dv_dy = np.gradient(comp_y, dx, dy)
    # swirling vectors matrix
    comp_x = dv_dx
    comp_y = du_dy
    # creating vectorfield object
    if raw:
        return (comp_x, comp_y)
    else:
        tmpvf = VectorField()
        unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
        unit_values = tmp_vf.unit_values
        tmpvf.import_from_arrays(axe_x, axe_y, comp_x, comp_y, mask=mask,
                                 unit_x=unit_x, unit_y=unit_y,
                                 unit_values=unit_values)
        return tmpvf


def get_swirling_vector(vf, raw=False):
    """
    Return a vector field with the swirling vectors
    (eigenvectors of the velocity laplacian matrix
    ponderated by eigenvalues)
    (Have to be adjusted : which part of eigenvalues
    and eigen vectors take ?)

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.
    """
    raise Warning("Useless (not finished)")
    if not isinstance(vf, VectorField):
        raise TypeError()
    tmp_vf = vf.copy()
    tmp_vf.fill(crop_border=True)
    # Getting gradients and axes
    axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
    comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    du_dx, du_dy = np.gradient(comp_x, dx, dy)
    dv_dx, dv_dy = np.gradient(comp_y, dx, dy)
    # swirling vectors matrix
    comp_x = np.zeros(tmp_vf.shape)
    comp_y = np.zeros(tmp_vf.shape)
    # loop on  points
    for i in np.arange(0, len(axe_x)):
        for j in np.arange(0, len(axe_y)):
            lapl = [[du_dx[i, j], du_dy[i, j]],
                    [dv_dx[i, j], dv_dy[i, j]]]
            eigvals, eigvect = np.linalg.eig(lapl)
            eigvals = np.imag(eigvals)
            eigvect = np.real(eigvect)
            if eigvals[0] > eigvals[1]:
                comp_x[i, j] = eigvect[0][0]
                comp_y[i, j] = eigvect[0][1]
            else:
                comp_x[i, j] = eigvect[1][0]
                comp_y[i, j] = eigvect[1][1]
    # creating vectorfield object
    if raw:
        return (comp_x, comp_y)
    else:
        unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
        # TODO: Implement unities
        unit_values = ""
        tmp_vf = VectorField()
        tmp_vf.import_from_arrays(axe_x, axe_y, comp_x, comp_y, mask=False,
                                  unit_x=unit_x, unit_y=unit_y,
                                  unit_values=unit_values)
        return tmp_vf


### Modifying field
def extrapolate_until_wall(field, direction='x', position=0.,
                           kind_interp='linear'):
    """
    Interpolate the given fiels until it reach  wall (0 velocity) at the given
    position.

    Parameters
    ----------
    field : ScalarField or VectorField object.
        Field to extend
    direction : string
        Axe on which the wall is placed (default to 'x')
    position : number
        position of the wall (in the same unit as the field)
    kind_interp : string
        Type of algorithm used to fill.
        ‘linear’ (default): fill using linear interpolation
        ‘cubic’ : fill using cubic interpolation

    Returns
    -------
    extended_field : ScalarField or VectorField
        Extended field.
    """
    # check params
    axe_x, axe_y = field.axe_x, field.axe_y
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    if not isinstance(direction, STRINGTYPES):
        raise TypeError()
    if not direction in ['x', 'y']:
        raise ValueError()
    if not isinstance(position, NUMBERTYPES):
        raise TypeError()
    if direction == 'x':
        if position < axe_x[-1] and position > axe_x[0]:
            raise ValueError()
    else:
        if position < axe_y[-1] and position > axe_y[0]:
            raise ValueError()
    if not isinstance(kind_interp, STRINGTYPES):
        raise TypeError()
    if not kind_interp in ['linear', 'cubic']:
        raise ValueError()
    # enlarge the field
    ext_field = field.copy()
    if direction == 'x':
        if position < axe_x[0]:
            ext_field.extend(nmb_left=np.int((axe_x[0] - position)/dx),
                             inplace=True)
        else:
            ext_field.extend(nmb_right=np.int((position - axe_x[1])/dx),
                             inplace=True)
    else:
        if position < axe_y[0]:
            ext_field.extend(nmb_down=np.int((axe_y[0] - position)/dy),
                             inplace=True)
        else:
            ext_field.extend(nmb_up=np.int((position - axe_y[1])/dy),
                             inplace=True)
    # ScalarField
    if isinstance(field, ScalarField):
        # adding wall
        if direction == 'x':
            if position < axe_x[0]:
                ext_field.values[0, :] = [0]*ext_field.shape[1]
                ext_field.mask[0, :] = [False]*ext_field.shape[1]
            else:
                ext_field.values[-1, :] = [0]*ext_field.shape[1]
                ext_field.mask[-1, :] = [False]*ext_field.shape[1]
        else:
            if position < axe_y[0]:
                ext_field.values[:, 0] = [0]*ext_field.shape[0]
                ext_field.mask[:, 0] = [False]*ext_field.shape[0]
            else:
                ext_field.values[:, -1] = [0]*ext_field.shape[0]
                ext_field.mask[:, -1] = [False]*ext_field.shape[0]
        # filling
        ext_field.fill(kind=kind_interp, inplace=True, reduce_tri=False)
        return ext_field
    # VectorField
    elif isinstance(field, VectorField):
        # adding wall
        if direction == 'x':
            if position < axe_x[0]:
                ext_field.comp_x[0, :] = [0]*ext_field.shape[1]
                ext_field.comp_y[0, :] = [0]*ext_field.shape[1]
                ext_field.mask[0, :] = [False]*ext_field.shape[1]
            else:
                ext_field.comp_x[-1, :] = [0]*ext_field.shape[1]
                ext_field.comp_y[-1, :] = [0]*ext_field.shape[1]
                ext_field.mask[-1, :] = [False]*ext_field.shape[1]
        else:
            if position < axe_y[0]:
                ext_field.comp_x[:, 0] = [0]*ext_field.shape[0]
                ext_field.comp_y[:, 0] = [0]*ext_field.shape[0]
                ext_field.mask[:, 0] = [False]*ext_field.shape[0]
            else:
                ext_field.comp_x[:, -1] = [0]*ext_field.shape[0]
                ext_field.comp_y[:, -1] = [0]*ext_field.shape[0]
                ext_field.mask[:, -1] = [False]*ext_field.shape[0]
        # filling
        ext_field.fill(kind=kind_interp, inplace=True, reduce_tri=False)
        return ext_field
    else:
        raise TypeError()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
try:
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(dirname)
    os.chdir(dirname)
except:
    pass
import matplotlib as mpl
mpl.use('Agg')

import numpy as np
import pytest
from helper import sane_parameters, parametric_test

import matplotlib.pyplot as plt
import unum
from IMTreatment import ScalarField, file_operation as imtio, make_unit


class TestScalarField(object):
    """ Done """

    def setup(self):
        sane_parameters()
        self.SF1 = imtio.import_from_file("SF1.cimt")
        self.SF1_nomask = imtio.import_from_file("SF1_nomask.cimt")
        self.SF2 = imtio.import_from_file("SF2.cimt")
        self.SF_notevenlyspaced = imtio.import_from_file("SF_notevenlyspaced.cimt")

    def test_import_from_arrays(self):
        # creating a SF field using 'import_from_arrays
        axe_x = np.arange(10)
        unit_x = make_unit('m')
        unit_y = make_unit('km')
        unit_values = make_unit('m/s')
        axe_y = np.arange(20)*.01
        values = np.arange(len(axe_x)*len(axe_y)).reshape(len(axe_y),
                                                          len(axe_x))
        values = np.array(values, dtype=float)
        mask = np.random.rand(len(axe_y), len(axe_x)) > 0.75
        values[mask] = np.nan
        values = np.pi*values
        sf = ScalarField()
        sf.import_from_arrays(axe_x, axe_y, values, mask=mask,
                              unit_x=unit_x, unit_y=unit_y,
                              unit_values=unit_values)
        values = values.transpose()
        mask = mask.transpose()
        # tests
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~sf.mask] == values[~mask])
        assert np.all(sf.mask == mask)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values

    def test_operations(self):
        # get datas
        axe_x, axe_y = self.SF1.axe_x, self.SF1.axe_y
        values = self.SF1.values
        mask = self.SF1.mask
        values2 = self.SF2.values
        mask2 = self.SF2.mask
        unit_x, unit_y = self.SF1.unit_x, self.SF1.unit_y
        unit_values = self.SF1.unit_values
        # neg
        sf = -self.SF1
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~sf.mask] == -values[~mask])
        assert np.all(sf.mask == mask)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # add
        nmb = 5
        unt = 500*make_unit('mm/s')
        values_f = (nmb + values + unt.asNumber()/1000. + values2 +
                    unt.asNumber()/1000. + values + nmb)
        mask_f = np.logical_or(mask, mask2)
        sf = nmb + self.SF1 + unt + self.SF2 + unt + self.SF1 + nmb
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~mask_f] == values_f[~mask_f])
        assert np.all(sf.mask == mask_f)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # sub
        nmb = 5
        unt = 500*make_unit('mm/s')
        values_f = (nmb - values - unt.asNumber()/1000. - values2 -
                    unt.asNumber()/1000. - values - nmb)
        mask_f = np.logical_or(mask, mask2)
        sf = nmb - self.SF1 - unt - self.SF2 - unt - self.SF1 - nmb
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~mask_f] == values_f[~mask_f])
        assert np.all(sf.mask == mask_f)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # mul
        nmb = 5.23
        unt = 500.*make_unit('mm/s')
        unt_n = 500./1000.
        values_f = (nmb * values * unt_n * values2 *
                    unt_n * values * nmb)
        unit_values = make_unit('mm/s')**2*make_unit('m/s')**3*1e6
        mask_f = np.logical_or(mask, mask2)
        sf = nmb * self.SF1 * unt * self.SF2 * unt * self.SF1 * nmb
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~mask_f] - values_f[~mask_f] < 1e-6)
        assert np.all(sf.mask == mask_f)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # div
        nmb = 5.23
        unt = 500.*make_unit('mm/s')
        unt_n = 500./1000.
        values_f = (nmb / values / unt_n / values2 /
                    unt_n / values / nmb)
        unit_values = 1./(make_unit('mm/s')**2*make_unit('m/s')**3*1e6)
        mask_f = np.logical_or(mask, mask2)
        sf = nmb / self.SF1 / unt / self.SF2 / unt / self.SF1 / nmb
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~mask_f] - values_f[~mask_f] < 1e-6)
        assert np.all(sf.mask == mask_f)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # abs
        unit_values = self.SF1.unit_values
        sf = np.abs(self.SF1)
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~sf.mask] == np.abs(values[~mask]))
        assert np.all(sf.mask == mask)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values
        # pow
        unit_values = self.SF1.unit_values**3.544186
        sf = (np.abs(self.SF1) + 1)**3.544186
        assert np.all(sf.axe_x == axe_x)
        assert np.all(sf.axe_y == axe_y)
        assert np.all(sf.values[~sf.mask] -
                      (np.abs(values[~mask]) + 1)**3.544186 <
                      1e-6)
        assert np.all(sf.mask == mask)
        assert sf.unit_x == unit_x
        assert sf.unit_y == unit_y
        assert sf.unit_values == unit_values

    def test_iter(self):
        axe_x, axe_y = self.SF1.axe_x, self.SF1.axe_y
        ind_x = np.arange(len(axe_x))
        ind_y = np.arange(len(axe_y))
        axe_x, axe_y = np.meshgrid(axe_x, axe_y)
        ind_x, ind_y = np.meshgrid(ind_x, ind_y)
        axe_x = np.transpose(axe_x)
        axe_y = np.transpose(axe_y)
        ind_x = np.transpose(ind_x)
        ind_y = np.transpose(ind_y)
        values = self.SF1.values
        mask = self.SF1.mask
        ind_x = ind_x[~mask]
        ind_y = ind_y[~mask]
        axe_x = axe_x[~mask]
        axe_y = axe_y[~mask]
        values = values[~mask]
        ind2_x = []
        ind2_y = []
        axe2_x = []
        axe2_y = []
        values2 = []
        for ij, xy, val in self.SF1:
            ind2_x.append(ij[0])
            ind2_y.append(ij[1])
            axe2_x.append(xy[0])
            axe2_y.append(xy[1])
            values2.append(val)
        assert np.all(ind_x == ind2_x)

    def test_trim_area(self):
        axe_x, axe_y = self.SF1.axe_x, self.SF1.axe_y
        values = self.SF1.values
        mask = self.SF1.mask
        sf = self.SF1.crop([axe_x[3], axe_x[-4]], [axe_y[2], axe_y[-7]])
        assert np.all(sf.axe_x == axe_x[3:-3])
        assert np.all(sf.axe_y == axe_y[2:-6])
        assert np.all(sf.values[~sf.mask] ==
                      values[3:-3, 2:-6][~mask[3:-3, 2:-6]])

    def test_min_max_mean(self):
        mini = self.SF1.min
        maxi = self.SF1.max
        mean = self.SF1.mean
        assert mini == -80.685138559926955
        assert maxi == -0.11918431073522018
        assert mean == -38.416196172428371

    def test_get_values(self):
        value = self.SF1.get_value(7.5, 20.12)
        value2 = self.SF1.get_value(5, 10, ind=True)
        assert value == -39.27599375418945
        assert value2 == -77.512615587040656

    def test_get_zones_centers(self):
        self.SF1.fill(inplace=True)
        zones_xy = np.genfromtxt("value_zones_center")
        zones = self.SF1.get_zones_centers(bornes=[0.75, 1])
        assert np.all(zones_xy == zones.xy)

    def test_get_nearest_extrema(self):
        self.SF1.fill(inplace=True)
        center = self.SF1.get_nearest_extrema([7.5, 20.12])
        assert (6.3020200717715031 - center[0][0]) < 1e-6
        assert (22.1880168513963919 - center[0][1]) < 1e-6

    def test_get_profile(self):
        profile = self.SF1.get_profile('x', 9.98)
        prof_x, prof_y = np.genfromtxt("profile")
        assert np.all(prof_x[~profile.mask] == profile.x[~profile.mask])
        assert np.all(prof_y[~profile.mask] == profile.y[~profile.mask])

    def test_get_spatial_autocorrelation(self):
        res_x = self.SF1.get_spatial_autocorrelation('x')
        # imtio.export_to_file(res_x, "spatial_autocorrelation_x.cimt")
        res_x2 = imtio.import_from_file("spatial_autocorrelation_x.cimt")
        assert res_x == res_x2
        res_y = self.SF1.get_spatial_autocorrelation('y')
        # imtio.export_to_file(res_y, "spatial_autocorrelation_y.cimt")
        res_y2 = imtio.import_from_file("spatial_autocorrelation_y.cimt")
        assert res_y == res_y2

    def test_get_spatial_spectrum(self):
        res_x = self.SF1.get_spatial_spectrum('x', intervx=[5, 19],
                                              intervy=[10.2, 29])
        # imtio.export_to_file(res_x, "get_spatial_spectrum_x.cimt")
        res_x2 = imtio.import_from_file("get_spatial_spectrum_x.cimt")
        assert res_x == res_x2
        res_y = self.SF1.get_spatial_spectrum('y', intervx=[5, 19],
                                              intervy=[10.2, 29])
        # imtio.export_to_file(res_y, "get_spatial_spectrum_y.cimt")
        res_y2 = imtio.import_from_file("get_spatial_spectrum_y.cimt")
        assert res_y == res_y2

    def test_get_norm(self):
        res = self.SF1.get_norm()
        # imtio.export_to_file(res, "get_norm.cimt")
        res2 = imtio.import_from_file("get_norm.cimt")
        assert res == res2

    def test_get_interpolator(self):
        res = self.SF1_nomask.get_interpolator()
        # imtio.export_to_file(res, "get_interpolator.cimt")
        res2 = imtio.import_from_file("get_interpolator.cimt")
        assert res(5, 9.2), res2(5 == 9.2)

    def test_integrate_over_line(self):
        res_x = self.SF1_nomask.integrate_over_line('x', [3.2, 17.8])
        # imtio.export_to_file(res_x, "integrate_over_line_x.cimt")
        res_x2 = imtio.import_from_file("integrate_over_line_x.cimt")
        assert res_x == res_x2
        res_y = self.SF1_nomask.integrate_over_line('y', [3.2, 17.8])
        # imtio.export_to_file(res_y, "integrate_over_line_y.cimt")
        res_y2 = imtio.import_from_file("integrate_over_line_y.cimt")
        assert res_y == res_y2

    def test_integrate_over_surface(self):
        res = self.SF1_nomask.integrate_over_surface(intervx=[2.4, 19.8],
                                                     intervy=[5.6, 8.3])
        # imtio.export_to_file(res, "integrate_over_surface.cimt")
        res2 = imtio.import_from_file("integrate_over_surface.cimt")
        assert res == res2

    def test_copy(self):
        res = self.SF1.copy()
        # imtio.export_to_file(res, "copy.cimt")
        res2 = imtio.import_from_file("copy.cimt")
        assert res == res2

    def test_export_to_scatter(self):
        res = self.SF1.export_to_scatter()
        # imtio.export_to_file(res, "export_to_scatter.cimt")
        res2 = imtio.import_from_file("export_to_scatter.cimt")
        assert res == res2

    def test_scale(self):
        res = self.SF1.scale(scalex=1.45, scaley=.98, scalev=3.4,
                             inplace=False)
        # imtio.export_to_file(res, "scale.cimt")
        res2 = imtio.import_from_file("scale.cimt")
        assert res == res2

    def test_rotate(self):
        res = self.SF1.rotate(270, inplace=False)
        # imtio.export_to_file(res, "rotate.cimt")
        res2 = imtio.import_from_file("rotate.cimt")
        assert res == res2

    def test_change_unit(self):
        res = self.SF1.copy()
        res.change_unit('x', 'um')
        res.change_unit('y', 'm')
        # imtio.export_to_file(res, "change_unit.cimt")
        res2 = imtio.import_from_file("change_unit.cimt")
        assert res == res2
        with pytest.raises(unum.IncompatibleUnitsError):
            res.change_unit('x', 'm/s')

    def test_crop(self):
        res = self.SF1.crop(intervx=[7.4, 19.0],
                            intervy=[2.1, 8], inplace=False)
        # imtio.export_to_file(res, "crop.cimt")
        res2 = imtio.import_from_file("crop.cimt")
        assert res == res2

    def test_extend(self):
        res = self.SF1.extend(nmb_left=1,
                              nmb_right=5,
                              nmb_down=7,
                              nmb_up=4,
                              inplace=False)
        # imtio.export_to_file(res, "extend.cimt")
        res2 = imtio.import_from_file("extend.cimt")
        assert res == res2

    def test_mirroring(self):
        res = self.SF1.mirroring('x', 35, mir_coef=1.5)
        # imtio.export_to_file(res, "mirroring.cimt")
        res2 = imtio.import_from_file("mirroring.cimt")
        assert res == res2
        #
        tmpsf = self.SF1.copy()
        tmpsf.mirroring('x', 35, mir_coef=1.5, inplace=True)
        res_b = tmpsf
        # imtio.export_to_file(res_b, "mirroring_b.cimt")
        res_b2 = imtio.import_from_file("mirroring_b.cimt")
        assert res_b == res_b2

    def test_crop_masked_border(self):
        tmpsf = self.SF1.copy()
        tmpsf.mask[0:3, :] = True
        tmpsf.mask[-1:, :] = True
        tmpsf.mask[:, 0:5] = True
        tmpsf.mask[:, -4:] = True
        res = tmpsf.crop_masked_border(inplace=False)
        # imtio.export_to_file(res, "crop_masked_border.cimt")
        res2 = imtio.import_from_file("crop_masked_border.cimt")
        assert res == res2
        #
        tmpsf = self.SF1_nomask.copy()
        tmpsf.mask[3:-2, :] = False
        tmpsf.mask[:, 5:-4] = False
        res = self.SF1.crop_masked_border(hard=True, inplace=False)
        # imtio.export_to_file(res, "crop_masked_border_2.cimt")
        res2 = imtio.import_from_file("crop_masked_border_2.cimt")
        assert res == res2

    def test_fill(self):
        res = self.SF1.fill(inplace=False)
        # imtio.export_to_file(res, "fill.cimt")
        res2 = imtio.import_from_file("fill.cimt")
        assert res == res2

    def test_smooth(self):
        res = self.SF1.smooth(tos='uniform', size=4, inplace=False)
        # imtio.export_to_file(res, "smooth.cimt")
        res2 = imtio.import_from_file("smooth.cimt")
        assert res == res2

    def test_make_evenly_spaced(self):
        res = self.SF_notevenlyspaced.make_evenly_spaced()
        # imtio.export_to_file(res, "make_evenly_spaced.cimt")
        res2 = imtio.import_from_file("make_evenly_spaced.cimt")
        assert res == res2

    def test_reduce_spatial_resolution(self):
        res = self.SF1.reduce_spatial_resolution(4, inplace=False)
        # imtio.export_to_file(res, "reduce_spatial_resolution.cimt")
        res2 = imtio.import_from_file("reduce_spatial_resolution.cimt")
        assert res == res2

    @pytest.mark.mpl_image_compare
    def test_display_SF_a(self):
        fig = plt.figure()
        self.SF1.display()
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_SF_b(self):
        fig = plt.figure()
        self.SF1.display(kind='contour')
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_SF_c(self):
        fig = plt.figure()
        self.SF1.display(kind='contourf')
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_SF_d(self):
        fig = plt.figure()
        self.SF1.display(kind='contourf', levels=[-1, 0, 1])
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_SF_e(self):
        fig = plt.figure()
        self.SF1.display('mask')
        return fig



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from ..core import VectorField, TemporalVectorFields, Points
from ..utils import make_unit
from ..utils.types import NUMBERTYPES, ARRAYTYPES
import copy
import matplotlib.pyplot as plt
from .. import boundary_layer as imtbl
from .. import potential_flow as pf
import scipy.interpolate as spinterp


class Vortex(object):
    """
    """

    def __init__(self, x0, y0, movable, idi=0):
        """

        """
        self.__x0 = x0
        self.__y0 = y0
        self.movable = movable
        self.rot_dir = 1
        self.id = idi

    @property
    def x0(self):
        return self.__x0

    @x0.setter
    def x0(self, new_x0):
        if not self.movable:
            raise Exception()
        self.__x0 = new_x0

    @property
    def y0(self):
        return self.__y0

    @y0.setter
    def y0(self, new_y0):
        if not self.movable:
            raise Exception()
        self.__y0 = new_y0

    def get_vector_field(self, axe_x, axe_y, unit_x='', unit_y=''):
        # preparing
        Vx = np.empty((len(axe_x), len(axe_y)), dtype=float)
        Vy = np.empty(Vx.shape, dtype=float)
        # getting velocities
        for i, x in enumerate(axe_x):
            for j, y in enumerate(axe_y):
                Vx[i, j], Vy[i, j] = self.get_vector(x, y)
        # returning
        vf = VectorField()
        vf.import_from_arrays(axe_x, axe_y, Vx, Vy, mask=False, unit_x=unit_x,
                              unit_y=unit_y, unit_values=make_unit('m/s'))
        return vf

    @staticmethod
    def _get_theta(x0, x, y0, y):
        """
        """
        dx = x - x0
        dy = y - y0
        if dx == 0 and dy == 0:
            theta = 0
        elif dx == 0:
            if dy > 0:
                theta = np.pi/2.
            else:
                theta = -np.pi/2.
        elif dx > 0.:
            theta = np.arctan((dy)/(dx))
        else:
            theta = np.arctan((dy)/(dx)) - np.pi
        return theta

    @staticmethod
    def _get_r(x0, x, y0, y):
        return ((x - x0)**2 + (y - y0)**2)**.5

    @staticmethod
    def _cyl_to_cart(theta, comp_r, comp_phi):
        c_theta = np.cos(theta)
        s_theta = np.sin(theta)
        comp_x = comp_r*c_theta - comp_phi*s_theta
        comp_y = comp_r*s_theta + comp_phi*c_theta
        return comp_x, comp_y

    def copy(self):
        return copy.deepcopy(self)

    def symetrize(self):
        """
        Return a symetrized copy of the vortex
        """
        tmp_vort = self.copy()
        tmp_vort.rot_dir *= -1
        return tmp_vort


class SolidVortex(Vortex):
    """
    Representing a solid rotation.
    """
    def __init__(self, x0=0., y0=0., omega=1., movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        omega : number, optional
            rotation velocity (rad/s)
        """
        super(SolidVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        if omega < 0:
            self.rot_dir = -1
        else:
            self.rot_dir = 1
        self.omega = np.abs(omega)

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in cylindrical referentiel
        Vr = 0.
        Vphi = r*self.omega*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class FreeVortex(Vortex):
    """
    Representing a Free (irrotational) Vortex.
    Due to its definition, the center of the vortex is a singular point
    (V = inf) set to 0 in this implementation.
    """
    def __init__(self, x0=0., y0=0., gamma=1., movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        gamma : number, optional
            Cirdculation of the free-vortex (m^2/s).
        """
        super(FreeVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        if gamma < 0:
            self.rot_dir = -1
        else:
            self.rot_dir = 1
        self.gamma = np.abs(gamma)

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in cylindrical referentiel
        Vr = 0.
        if r == 0:
            Vphi = 0
        else:
            Vphi = self.gamma/(2*np.pi*r)*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class BurgerVortex(Vortex):
    """
    Representing a Burger Vortex, a stationnary self-similar flow, caused by
    the balance between vorticity creation at the center and vorticity
    diffusion.

    Notes
    -----
    Analytical Vortex Solutions to the Navier-Stokes Equation.
    Thesis for the degree of Doctor of Philosophy, Växjö University,
    Sweden 2007.
    """
    def __init__(self, x0=0., y0=0., alpha=1e-6, ksi=1., viscosity=1e-6,
                 movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        alpha : number, optional
            Positive constant (default : 1e-6), determine te size of the
            vortex. To have a vortex of radius 'R', 'alpha' should be equal to
            '9.2*viscosity/R**2'.
        ksi : number, optional
            Constant (default : 1.), make the overall velocity augment.
            Can also be used to switch the rotation direction.
        viscosity : number, optional
            Viscosity (default : 1e-6 (water))
        """
        super(BurgerVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        if ksi < 0:
            self.rot_dir = -1
        else:
            self.rot_dir = 1
        self.ksi = np.abs(ksi)
        self.alpha = alpha
        self.viscosity = viscosity
        self.C1 = -self.alpha/(4.*self.viscosity)
        self.C2 = self.ksi/(2*np.pi)
        self.C3 = -1/2.*self.alpha

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in clyndrical referentiel
        Vr = self.C3*r
        if r == 0:
            Vphi = 0
        else:
            Vphi = (self.C2/r) \
                * (1 - np.exp(self.C1*r**2))*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class HillVortex(Vortex):
    """
    Representing a Hill Vortex, a convected vortex sphere in a inviscid flow.

    Notes
    -----
    Analytical Vortex Solutions to the Navier-Stokes Equation.
    Thesis for the degree of Doctor of Philosophy, Växjö University,
    Sweden 2007.
    """
    def __init__(self, x0=0, y0=0, U=1., rv=1., rot_dir=1, unit_values='',
                 movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        U : number
            Convection velocity (m/s)
        rv : number
            Vortex radius
        """
        super(HillVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        self.U = U
        self.rv = rv
        self.rot_dir = rot_dir

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in clyndrical referentiel
        Vr = -3./4.*self.U*r*(1 - r**2/self.rv**2) \
            * 2*np.sin(theta)*np.cos(theta)
        Vphi = 3./2.*self.U*np.sin(theta)**2 \
            * r*(1 - 2*r**2/self.rv**2)*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class LambOseenVortex(Vortex):
    """
    Representing a Lamb-Oseen Vortex, a vortex with decay due to viscosity.
    (satisfy NS)

    Notes
    -----
    Analytical Vortex Solutions to the Navier-Stokes Equation.
    Thesis for the degree of Doctor of Philosophy, Växjö University,
    Sweden 2007.
    """
    def __init__(self, x0=0, y0=0, ksi=1., t=1., viscosity=1e-6, movable=True,
                 idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        ksi : number
            Overall velocity factor.
        t : number
            Time.
        viscosity : number
            Viscosity (default : 1e-6 (water))
        """
        super(LambOseenVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        if ksi < 0:
            self.rot_dir = -1
        else:
            self.rot_dir = 1
        self.ksi = np.abs(ksi)
        self.t = t
        self.viscosity = viscosity

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in clyndrical referentiel
        Vr = 0.
        if r == 0:
            Vphi = 0
        else:
            Vphi = self.ksi/(2*np.pi*r)\
                * (1 - np.exp(-r**2/(4.*self.viscosity*self.t)))*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class RankineVortex(Vortex):
    """
    Representing a Rankine Vortex, with an inner zone or forced vortex, and
    an outer zone of free vortex.

    Notes
    -----
    Giaiotti, DARIO B., et FULVIO Stel. « The Rankine vortex model ».
    PhD course on Environmental Fluid Mechanics-ICTP/University of Trieste,
    2006.

    """
    def __init__(self, x0=0., y0=0., circ=1., rv=1., movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        rv : number
            Vortex inner zone radius
        circ : number
            Vortex circulation (m^2/s)
        unit_values : string
            Velocity unity
        """
        super(RankineVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        self.rv = rv
        if circ < 0:
            self.rot_dir = -1
        else:
            self.rot_dir = 1
        self.circ = np.abs(circ)

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in clyndrical referentiel
        Vr = 0.
        if r <= self.rv:
            Vphi = self.circ*r/(2*np.pi*self.rv**2)*self.rot_dir
        else:
            Vphi = self.circ/(2*np.pi*r)*self.rot_dir
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy


class LambChaplyginVortex(Vortex):
    """
    Representing a Lamb-Chaplygin dipole vortex, with potential flow in the
    exterior region, and a linear relation between stream function and
    vorticity in the inner region.

    Notes
    -----
    Analytical Vortex Solutions to the Navier-Stokes Equation.
    Thesis for the degree of Doctor of Philosophy, Växjö University,
    Sweden 2007.
    """
    def __init__(self, x0=0, y0=0, U=1., rv=1., Bessel_root_nmb=1,
                 movable=True, idi=0):
        """
        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        U : number
            Convection velocity (m/s)
        rv : number
            Delimitation radius between interior and exterior (m).
        Bessel_root_nmb : integer
            Bessel root number evaluated to choose the constant k

        """
        super(LambChaplyginVortex, self).__init__(x0, y0, movable=movable,
                                                  idi=idi)
        self.U = U
        self.rv = rv
        self.Bessel_root_nmb = Bessel_root_nmb

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        from scipy.special import jn, jn_zeros
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute Bessel root number
        k = jn_zeros(1, self.Bessel_root_nmb)[-1]/self.rv
        # compute stream function
        if r < self.rv:
            J1_p = (jn(2, k*r) - jn(0, k*r))/(-2)
            Vr = -2*self.U/(r*k*jn(0, k*self.rv))*jn(1, k*r)*np.cos(theta)
            Vphi = 2*self.U/(jn(0, k*self.rv))*np.sin(theta)*J1_p
        else:
            Vr = -self.U*(r - self.rv**2/r)*np.cos(theta)
            Vphi = +self.U*(1 + self.rv**2/r**2)*np.sin(theta)
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy

    def J(self, order, x):
        """
        Return the value of the Bessel function with the given order ar the
        given point.

        Parameters
        ----------
        order : number
            Order of the Bessel function
        x : number
            Value where we want the Bessel function evaluation.

        Returns
        -------
        y : number
            Bessel function value at 'x'
        """
        from scipy.special import jn
        return jn(order, x)


class PersoVortex(Vortex):
    """
    Representing a slice of a 3D vortex.
    """
    def __init__(self, x0=0, y0=0, radius=1., vort_max=None, circ=None,
                 node_ratio=0., nu=1e-6, movable=True, idi=0):
        """
        Representing a slice of a 3D vortex.

        Notes
        -----
        Modified Burger vortex :

        .. math::
            u_{\\theta} = \\frac{\\Gamma}{2\\pi r} \\left[ 1 - \\exp
            \\left( -\\frac{\\alpha r^2}{4 \\nu} \\right) \\right]

            u_r = -\\frac{\\alpha r}{2}

            u_z = \\alpha z

        respectant les propriétés suivantes (vorticité et rayon à i%) :

        .. math::
            \\omega = \\frac{\\alpha \\Gamma}{4 \\pi \\nu} \\exp \\left(
            -\\frac{\\alpha r^2}{4 \\nu} \\right)

            R_{i} = \\left( -\\frac{16 \\ln(i)\\nu}{\\alpha} \\right)^{0.5}

        Parameters
        ----------
        x0, y0 : numbers, optional
            Position of the vortex center (default : [0, 0]).
        """
        # check
        alpha = nu/(radius/2)**2
        alpha = 16*np.log(2)*nu/radius**2
        if circ is None:
            if vort_max is None:
                raise ValueError()
            circ = 2*vort_max*4*np.pi*nu/(alpha)
        # store
        super(PersoVortex, self).__init__(x0, y0, movable=movable, idi=idi)
        self.radius = radius
        self.circ = circ
        self.node_ratio = node_ratio
        self.nu = nu
        self.alpha = alpha
        # compute vortex constant (for fast velocity computation)
        self.C1 = -self.alpha/(4.*self.nu)
        self.C2 = self.circ/(2*np.pi)
        self.C3 = -1/2.*self.alpha

    def get_vector(self, x, y):
        """
        Return the velocity vector at the given point.
        """
        # compute r
        r = self._get_r(self.x0, x, self.y0, y)
        # compute theta
        theta = self._get_theta(self.x0, x, self.y0, y)
        # compute velocity in clyndrical referentiel
        if r == 0:
            Vphi = 0
        else:
            Vphi = (self.C2/r) \
                * (1 - np.exp(self.C1*r**2))*self.rot_dir
        Vr = -np.abs(Vphi)*self.node_ratio
        # get velocity in the cartesian refenrentiel
        Vx, Vy = self._cyl_to_cart(theta, Vr, Vphi)
        # returning
        return Vx, Vy

class CustomField(object):
    """
    Representing a custom field.

    Parameters
    ----------
    funct : function
        Representing the field.
        Has to take (x, y) as input and return (Vx, Vy).
    """
    def __init__(self, funct, unit_values='', idi=0):
        # check params
        try:
            Vx, Vy = funct(1., 1.)
        except TypeError:
            raise TypeError()
        else:
            if (not isinstance(Vx, NUMBERTYPES)
                    or not isinstance(Vy, NUMBERTYPES)):
                raise TypeError()
        # store
        self.funct = funct
        self.unit_values = unit_values
        self.id = idi

    def copy(self):
        """
        """
        return copy.deepcopy(self)

    def get_vector(self, x, y):
        """
        """
        return self.funct(x, y)

    def get_vector_field(self, axe_x, axe_y, unit_x='', unit_y=''):
        # preparing
        Vx = np.empty((len(axe_x), len(axe_y)), dtype=float)
        Vy = np.empty(Vx.shape, dtype=float)
        mask = np.empty(Vx.shape, dtype=bool)
        # getting velocities
        for i, x in enumerate(axe_x):
            for j, y in enumerate(axe_y):
                Vx[i, j], Vy[i, j] = self.funct(x, y)
        mask = np.logical_or(np.isnan(Vx), np.isnan(Vy))
        # returning
        vf = VectorField()
        vf.import_from_arrays(axe_x, axe_y, Vx, Vy, mask=mask, unit_x=unit_x,
                              unit_y=unit_y, unit_values=self.unit_values)
        return vf


class Wall(object):
    """
    Representing a wall
    """
    def __init__(self, x=None, y=None):
        """
        Representing a wall

        Parameters
        ----------
        x, y : numbers
            Position(s) of the wall along 'x' and 'y'.
        """
        # check
        if x is not None:
            self.direction = 'x'
            self.position = x
            if y is not None:
                raise ValueError("'x' and 'y' cannot be both defined")
        elif y is not None:
            self.direction = 'y'
            self.position = y
        else:
            raise ValueError("'x' or 'y' should be defined")

    def get_symmetry(self, pt):
        """
        Give the symmetry of 'pt' according to the wall.
        """
        # check
        if not isinstance(pt, ARRAYTYPES):
            raise TypeError()
        pt = np.array(pt)
        if not pt.shape == (2,):
            raise ValueError()
        # get symmetry
        if self.direction == 'x':
            new_x = self.position + (self.position - pt[0])
            return [new_x, pt[1]]
        else:
            new_y = self.position + (self.position - pt[1])
            return [pt[0], new_y]


class VortexSystem(object):
    """
    Representing a set of vortex.
    """

    def __init__(self):
        """
        Representing a set of vortex.
        """
        self.vortex = []
        self.im_vortex = []
        self.nmb_vortex = 0
        self.walls = []
        self.custfields = []
        self.to_refresh = False
        self.curr_id = 0
        self.curr_cf_id = 0

    def copy(self):
        """
        Return a copy.
        """
        return copy.deepcopy(self)

    def add_vortex(self, vortex):
        """
        Add a vortex, or a custom field to the set.

        Parameters
        ----------
        vortex : Vortex or CustomField object
            vortex or field to add to the set
        """
        if not isinstance(vortex, Vortex):
            raise TypeError()
        vort = vortex.copy()
        vort.id = self.curr_id
        self.curr_id += 1
        self.vortex.append(vort)
        self.nmb_vortex += 1
        self.to_refresh = True

    def add_wall(self, wall):
        """
        Add a wall to the vortex system

        Parameters
        ----------
        wall : Wall object
            Wall to add.
        """
        if not isinstance(wall, Wall):
            raise TypeError()
        self.walls.append(wall)
        self.to_refresh = True

    def add_custom_field(self, custfield):
        """
        Add a custom field to the vortex system

        Parameters
        ----------
        custfield : CustomField object
            Custom field to add.
        """
        if not isinstance(custfield, CustomField):
            raise TypeError()
        custfield.id = self.curr_cf_id
        self.curr_cf_id += 1
        self.custfields.append(custfield)

    def remove_vortex(self, ind):
        """
        Remove a vortex or a custom field from the set.

        Parameters
        ----------
        ind : integer
            Vortex indice to remove.
        """
        if ind < 0:
            ind = len(self.vortex) + ind
        id_to_remove = self.vortex[ind].id
        # remove vortex
        for i in np.arange(len(self.vortex) - 1, -1, -1):
            vort = self.vortex[i]
            if vort.id == id_to_remove:
                self.vortex.pop(i)
            elif vort.id > id_to_remove:
                self.vortex[i].id -= 1
        # remove imaginary vortex
        for i in np.arange(len(self.im_vortex) - 1, -1, -1):
            vort = self.im_vortex[i]
            if vort.id == id_to_remove:
                self.im_vortex.pop(i)
            elif vort.id > id_to_remove:
                self.im_vortex[i].id -= 1
        # adapt current id
        self.curr_id -= 1

    def display(self):
        """
        Display a representation of the vortex system
        """
        # define vortx centers colors
        colors = [plt.cm.jet(i) for i in np.linspace(0, 1, len(self.vortex))]

        # refresh if necessary
        if self.to_refresh:
            self.refresh_imaginary_vortex()
        # display vortex
        for vort in self.vortex:
            color = colors[vort.id]
            plt.plot(vort.x0, vort.y0, marker='o', mec='w', mfc=color)
        # display walls
        for wall in self.walls:
            if wall.direction == 'x':
                plt.axvline(wall.position, color='k')
            else:
                plt.axhline(wall.position, color='k')
        # display imaginary vortex
        for ivort in self.im_vortex:
            color = colors[ivort.id]
            plt.plot(ivort.x0, ivort.y0, marker='o', mec='w', mfc=color)

    def display_compounded_vector(self, x, y, scale=1., detailed=False):
        """
        Display the compounded vector in 'x', 'y' position.
        """
        # get data
        colors = [plt.cm.jet(i) for i in np.linspace(0, 1, len(self.vortex))]
        vort_vects, vort_ids, cf_vects, cf_ids \
            = self.get_compounded_vector(x, y, detailed=detailed)
        norm = plt.Normalize(0, len(self.vortex))
        # loop on vortex vectors
        for i, vect in enumerate(vort_vects):
            idi = vort_ids[i]
            if i >= len(colors):
                color = 'k'
            else:
                color = colors[i]
            plt.quiver(x, y, *vect, color=color, norm=norm, scale=scale)
        # loop on custom fields
        for i, vect in enumerate(cf_vects):
            color = 'k'
            plt.quiver(x, y, *vect, color=color, norm=norm, scale=scale)

    def get_vector(self, x, y, vortex_ids='all', cf_ids='all'):
        """
        Return the resulting velocity vector, at the given point.

        Parameters
        ----------
        x, y : numbers
            Position of the wanted vector.
        vortex_ids : list of integers, 'all' or 'none'
            Can be used to filter the wanted vortex influence by ids.
        cf_ids : list of integers, 'all' or 'none'
            Can be used to filter the wanted custom fields influence by ids.
        Returns
        -------
        Vx, Vy : numbers
            Velocity components.
        """
        # check
        if vortex_ids == 'all':
            vortex_ids = np.arange(self.curr_id)
        elif vortex_ids == 'none':
            vortex_ids = []
        if cf_ids == 'all':
            cf_ids = np.arange(self.curr_cf_id)
        elif cf_ids == 'none':
            cf_ids = []
        # refesh if necessary
        if self.to_refresh:
            self.refresh_imaginary_vortex()
        # create vector
        Vx = 0.
        Vy = 0.
        # add vortex participation
        for vort in self.vortex:
            if vort.id not in vortex_ids:
                continue
            tmp_Vx, tmp_Vy = vort.get_vector(x, y)
            Vx += tmp_Vx
            Vy += tmp_Vy
        # add imaginary vortex participation
        for vort in self.im_vortex:
            if vort.id not in vortex_ids:
                continue
            tmp_Vx, tmp_Vy = vort.get_vector(x, y)
            Vx += tmp_Vx
            Vy += tmp_Vy
        # add custom fields
        for cst_field in self.custfields:
            if cst_field.id not in cf_ids:
                continue
            tmp_Vx, tmp_Vy = cst_field.get_vector(x, y)
            Vx += tmp_Vx
            Vy += tmp_Vy
        # returning
        return Vx, Vy

    def get_compounded_vector(self, x, y, detailed=False):
        """
        Return a list f vector, for each structure of the system

        Parameters
        ----------
        x, y : numbers
            Wanted vector coordinates
        detailed : boolean, optional
            If 'False' (default), vortex contribution and its imaginary vortex
            contribution are summed up. Else, return one vector for each
            contribution.

        Returns
        -------
        vects : Nx2 array of numbers
            Vectors associated to vortex contribution
        vortex_ids : Nx1 array of integer
            Associated id
        verts_cf : Mx2 array of numbers
            Vectors associated to custom fields contribution
        cf_ids : Mx1 array of integer
            Associated id
        """
        vects = []
        vects_cf = []
        if detailed:
            vortex_ids = []
            cf_ids = []
            # loop on vortex
            for vort in self.vortex:
                vects.append(vort.get_vector(x, y))
                vortex_ids.append(vort.id)
            # loop on imaginary vortex
            for vort in self.im_vortex:
                vects.append(vort.get_vector(x, y))
                vortex_ids.append(vort.id)
            # loop on custom field
            for cf in self.custfields:
                vects_cf.append(cf.get_vector(x, y))
                cf_ids.append(cf.id)
        else:
            vortex_ids = np.arange(self.curr_id)
            cf_ids = np.arange(self.curr_cf_id)
            # loop on vortex ids
            for idi in vortex_ids:
                vects.append(self.get_vector(x, y, vortex_ids=[idi],
                                             cf_ids='none'))
            # loop on cf ids
            for idi in cf_ids:
                vects_cf.append(self.get_vector(x, y, vortex_ids='none',
                                                cf_ids=[idi]))
        # return
        return vects, vortex_ids, vects_cf, cf_ids

    def refresh_imaginary_vortex(self):
        """
        """
        self.im_vortex = []
        for wall in self.walls:
            for i, ivort in enumerate(self.im_vortex[:]):
                im_vort = ivort.symetrize()
                im_vort.movable = True
                im_vort.x0, im_vort.y0 = wall.get_symmetry((ivort.x0,
                                                            ivort.y0))
                self.im_vortex.append(im_vort)
            for i, vort in enumerate(self.vortex):
                im_vort = vort.symetrize()
                im_vort.movable = True
                im_vort.x0, im_vort.y0 = wall.get_symmetry((vort.x0, vort.y0))
                self.im_vortex.append(im_vort)
        self.to_refresh = False

    def get_vector_field(self, axe_x, axe_y, unit_x='', unit_y=''):
        """
        Return a vector field on the given grid

        Parameters
        ----------
        axe_x, axe_y : arrays of crescent numbers
            x and y axis
        unit_x, unit_y : string or Unum objects
            Axis unities

        Returns
        -------
        vf :  VectorField object
            .
        """
        # check
        if not isinstance(axe_x, ARRAYTYPES):
            raise TypeError()
        if not isinstance(axe_y, ARRAYTYPES):
            raise TypeError()
        # refresh if necessary
        if self.to_refresh:
            self.refresh_imaginary_vortex()
        # get data
        axe_x = np.array(axe_x)
        axe_y = np.array(axe_y)
        vx = np.zeros((len(axe_x), len(axe_y)))
        vy = vx.copy()
        VF = VectorField()
        VF.import_from_arrays(axe_x, axe_y, vx, vy, mask=False, unit_x=unit_x,
                              unit_y=unit_y, unit_values='m/s')
        # add vortex participation
        for vort in self.vortex:
            VF += vort.get_vector_field(axe_x, axe_y,
                                        unit_x=unit_x, unit_y=unit_y)
        # add imaginary vortex participation
        for ivort in self.im_vortex:
            VF += ivort.get_vector_field(axe_x, axe_y,
                                         unit_x=unit_x, unit_y=unit_y)
        # add custom fields
        for cst_field in self.custfields:
            VF += cst_field.get_vector_field(axe_x, axe_y,
                                             unit_x=unit_x, unit_y=unit_y)
        # returning
        return VF

    def get_evolution(self, dt=1.):
        """
        Change the position of the vortex, according to the resulting velocity
        field and the time step.

        Parameters
        ----------
        dt : number
            time step.

        Returns
        -------
        vs : VortexSystem object
            New vortex system at t+dt

        """
        # check
        if not isinstance(dt, NUMBERTYPES):
            raise TypeError()
        if dt <= 0:
            raise ValueError()
        new_vs = self.copy()
        # loop on vortex
        for i, vort in enumerate(self.vortex):
            if not vort.movable:
                continue
            # get velocity on the vortex core
            Vx, Vy = self.get_vector(vort.x0, vort.y0)
            # get the vortex core dispacement
            dx, dy = dt*Vx, dt*Vy
            # change the vortex position in the new vortex system
            new_vs.vortex[i].x0 += dx
            new_vs.vortex[i].y0 += dy
        new_vs.refresh_imaginary_vortex()
        # returning
        return new_vs

    def get_temporal_vector_field(self, dt, axe_x, axe_y, nmb_it, unit_x='',
                                  unit_y='', unit_time=''):
        """
        """
        # create tvf
        time = 0.
        tmp_tvf = TemporalVectorFields()
        tmp_vf = self.get_vector_field(axe_x, axe_y, unit_x=unit_x,
                                       unit_y=unit_y)
        tmp_tvf.add_field(tmp_vf, time=time, unit_times=unit_time)
        time += dt
        # make time iterations
        tmp_vs = self.copy()
        for i in np.arange(nmb_it):
            tmp_vs = tmp_vs.get_evolution(dt=dt)
            tmp_vf = tmp_vs.get_vector_field(axe_x, axe_y, unit_x=unit_x,
                                             unit_y=unit_y)
            tmp_tvf.add_field(tmp_vf, time=time, unit_times=unit_time)
            time += dt
        # returning
        return tmp_tvf

    def get_vortex_evolution(self, dt, nmb_it, unit_time=''):
        """
        """
        time = 0.
        # create storage
        tmp_pts = []
        for vort in self.vortex:
            tmp_pts.append(Points(xy=[[vort.x0, vort.y0]], v=[time],
                                  unit_v=unit_time))
        # make time iterations
        tmp_vs = self.copy()
        for i in np.arange(nmb_it):
            tmp_vs = tmp_vs.get_evolution(dt=dt)
            # add vortex positions
            for i, vort in enumerate(tmp_vs.vortex):
                tmp_pts[i].add([vort.x0, vort.y0], v=time + dt)
            time += dt
        # returning
        return tmp_pts

    def get_velocity_map(self, axe_x, axe_y, vort, unit_x='m', unit_y='m'):
        """
        Return the velocity map of vortex 'vort' on the vortex system
        (velocity that the vortex 'vort' will have if placed at different
        space position on the vortex system)
        """
        # prepare
        comp_x = np.zeros((len(axe_x), len(axe_y)))
        comp_y = np.zeros(comp_x.shape)
        tmp_vs = self.copy()
        tmp_vort = vort.copy()
        # loop on frid points
        for i, x in enumerate(axe_x):
            for j, y in enumerate(axe_y):
                tmp_vort.x0 = x
                tmp_vort.y0 = y
                tmp_vs.add_vortex(tmp_vort)
                V = tmp_vs.get_vector(x, y)
                comp_x[i, j] = V[0]
                comp_y[i, j] = V[1]
                tmp_vs.remove_vortex(-1)
        # return
        tmp_vf = VectorField()
        tmp_vf.import_from_arrays(axe_x=axe_x, axe_y=axe_y, comp_x=comp_x,
                                  comp_y=comp_y, mask=False,
                                  unit_x=unit_x, unit_y=unit_y,
                                  unit_values='m/s')
        return tmp_vf


class HsvSystem(VortexSystem):

    def __init__(self, u_D, D, h, delta, vertical_wall=True):
        super(HsvSystem, self).__init__()
        # store data
        self.u_D = u_D
        self.D = D
        self.h = h
        self.delta = delta
        # add walls
        if vertical_wall:
            self.add_wall(Wall(x=0))
        self.add_wall(Wall(y=0))
        # add bg_field
        bl_funct = self._get_bl_profile(u_D, delta, h)
        sd_funct = self._get_velocity_slowdown(D)

        def bg_field(x, y):
            Vx = bl_funct(y)*sd_funct(x)
            return Vx, 0
        self.add_custom_field(CustomField(bg_field, unit_values='m/s'))
        #

    @staticmethod
    def _get_bl_profile(u_D, delta, h):
        blas_bl = imtbl.BlasiusBL(u_D, 1e-6, 1000.)
        x_prof = blas_bl.get_x_from_delta(delta)
        prof = blas_bl.get_profile(x=x_prof, y=np.linspace(0, h, 100))
        bl_funct = spinterp.interp1d(prof.x, prof.y, kind='linear', axis=0,
                                     copy=True, bounds_error=False,
                                     fill_value=0)
        return bl_funct

    @staticmethod
    def _get_velocity_slowdown(D):
        # create potential flow
        PS = pf.System(u_inf=1., alpha=0.)
        PS.add_object(2, [[0, -D/2], [0, D/2], [D, D/2], [D, -D/2]], res=20)
        PS.objects_2D[0].inverse_normals()
        Vx, _ = PS.compute_velocity_on_line([-5*D, 0], [0, 0], res=200,
                                            raw=False, remove_solid=False)
        Vx.x -= 5*D
        sd_funct = spinterp.interp1d(Vx.x, Vx.y, kind='linear', axis=0,
                                     copy=True, bounds_error=False,
                                     fill_value=0)
        return sd_funct


class StepSystem(VortexSystem):

    def __init__(self, u_D, H, h, delta):
        super(StepSystem, self).__init__()
        # store data
        self.u_D = u_D
        self.h = h
        self.H = H
        self.delta = delta
        # add walls
        self.add_wall(Wall(x=0))
        self.add_wall(Wall(y=0))
        # add bg_field
        bl_funct = self._get_bl_profile(delta, h, u_D)
        dev_funct = self._get_velocity_deviation(H, h, u_D)

        def bg_field(x, y):
            V = dev_funct(x, y)
            V[0] *= bl_funct(y)
            bl2 = bl_funct(-x)  # TEMPORARY
            V[1] *= bl2         # TEMPORARY
            V[0] *= bl2         # TEMPORARY
            return V
        self.add_custom_field(CustomField(bg_field, unit_values='m/s'))

    @staticmethod
    def _get_bl_profile(delta, h, u_D):
        blas_bl = imtbl.BlasiusBL(u_D, 1e-6, 1000.)
        x_prof = blas_bl.get_x_from_delta(delta)
        prof = blas_bl.get_profile(x=x_prof, y=np.linspace(0, h, 100))
        bl_funct = spinterp.interp1d(prof.x, prof.y/u_D, kind='linear', axis=0,
                                     copy=True, bounds_error=False,
                                     fill_value=0)
        return bl_funct

    @staticmethod
    def _get_velocity_deviation(H, h, u_D):
        # create potential flow
        PS = pf.System(u_inf=u_D, alpha=0.)
        PS.add_object(2, [[1.*H, -H], [0, -H], [0, H], [1.*H, H]], res=20)
        PS.objects_2D[0].inverse_normals()
        axe_x = np.linspace(-3*H, 0, 31)
        axe_y = np.arange(0, h + H/10., H/10.)
        VF = PS.compute_velocity_on_grid(axe_x, axe_y, remove_solid=True)
        interp_x = spinterp.interp2d(axe_x, axe_y, np.transpose(VF.comp_x),
                                     fill_value=1.)
        interp_y = spinterp.interp2d(axe_x, axe_y, np.transpose(VF.comp_y),
                                     fill_value=0)

        def dev_funct(x, y):
            return [interp_x(x, y)[0], interp_y(x, y)[0]]
        return dev_funct



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
try:
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(dirname)
    os.chdir(dirname)
except:
    pass
import matplotlib as mpl
mpl.use('Agg')


import numpy as np
import pytest
from helper import parametric_test, sane_parameters

import matplotlib.pyplot as plt
import unum

from IMTreatment import VectorField, file_operation as imtio, make_unit, \
    TemporalVectorFields
import IMTreatment.vortex_detection as vod
import IMTreatment.vortex_criterions as voc


class TestVortexDetection(object):

    def setup(self):
        sane_parameters()
        # VF1 = VectorField()
        # VF1.import_from_arrays(np.linspace(0, 24, 40),
        #                        np.linspace(-15.5, 15.5, 40),
        #                        (np.random.rand(40, 40) - .5)*12.3,
        #                        (np.random.rand(40, 40) - .5)*4.35,
        #                        mask=False,
        #                        unit_x='m',
        #                        unit_y='m',
        #                        unit_values='m/s')
        # VF1.smooth(tos='gaussian', size=2, inplace=True)
        # TVF1 = TemporalVectorFields()
        # for i in range(10):
        #     TVF1.add_field(VF1*np.cos(i/5*np.pi), i*0.834)
        # # imtio.export_to_file(VF1, 'VF1_nomask.cimt')
        # # imtio.export_to_file(TVF1, 'TVF1_nomask.cimt')
        self.VF1_nomask = imtio.import_from_file('VF1_nomask.cimt')
        self.TVF1_nomask = imtio.import_from_file('TVF1_nomask.cimt')

    def test_get_critical_points_a(self):
        res_a = vod.get_critical_points(self.VF1_nomask,
                                        mirroring=[['x', 0.]],
                                        thread='all')
        # imtio.export_to_file(res_a, "VF1_get_critical_points_a.cimt")
        res_a2 = imtio.import_from_file("VF1_get_critical_points_a.cimt")
        assert len(res_a2.foc) > 0
        assert res_a == res_a2

    def test_get_critical_points_b(self):
        res_b = vod.get_critical_points(self.TVF1_nomask)
        # imtio.export_to_file(res_b, "VF1_get_critical_points_b.cimt")
        res_b2 = imtio.import_from_file("VF1_get_critical_points_b.cimt")
        assert len(res_b2.foc) > 0
        assert res_b == res_b2

    def test_get_critical_points_c(self):
        res_c = vod.get_critical_points(self.TVF1_nomask, kind='gam_vort',
                                        thread='all')
        # imtio.export_to_file(res_c, "VF1_get_critical_points_c.cimt")
        res_c2 = imtio.import_from_file("VF1_get_critical_points_c.cimt")
        assert len(res_c2.foc) > 0
        assert res_c == res_c2

    def test_get_critical_points_d(self):
        res_d = vod.get_critical_points(self.VF1_nomask, kind='gam_vort',
                                        mirroring=[['x', 0.]])
        # imtio.export_to_file(res_d, "VF1_get_dritical_points_d.cimt")
        res_d2 = imtio.import_from_file("VF1_get_dritical_points_d.cimt")
        assert len(res_d2.foc) > 0
        assert res_d == res_d2

    def test_get_critical_points_e(self):
        res_e = vod.get_critical_points(self.TVF1_nomask, kind='pbi_cell')
        # imtio.export_to_file(res_e, "VF1_get_eritical_points_e.cimt")
        res_e2 = imtio.import_from_file("VF1_get_eritical_points_e.cimt")
        assert len(res_e2.foc) > 0
        assert res_e == res_e2

    def test_get_critical_points_f(self):
        res_f = vod.get_critical_points(self.VF1_nomask, kind='pbi_cell',
                                        mirroring=[['x', 0.]])
        # imtio.export_to_file(res_f, "VF1_get_fritical_points_f.cimt")
        res_f2 = imtio.import_from_file("VF1_get_fritical_points_f.cimt")
        assert len(res_f2.foc) > 0
        assert res_f == res_f2

    def test_get_critical_points_g(self):
        res_g = vod.get_critical_points(self.TVF1_nomask, kind='pbi_crit',
                                        thread='all')
        # imtio.export_to_file(res_g, "VF1_get_gritical_points_g.cimt")
        res_g2 = imtio.import_from_file("VF1_get_gritical_points_g.cimt")
        assert len(res_g2.foc) > 0
        assert res_g == res_g2

    def test_get_critical_points_h(self):
        res_h = vod.get_critical_points(self.VF1_nomask, kind='pbi_crit',
                                        smoothing_size=2)
        # imtio.export_to_file(res_h, "VF1_get_hritical_points_h.cimt")
        res_h2 = imtio.import_from_file("VF1_get_hritical_points_h.cimt")
        assert len(res_h2.foc) > 0
        assert res_h == res_h2

    def test_compute_traj(self):
        res_g = vod.get_critical_points(self.TVF1_nomask, kind='pbi_crit',
                                        thread='all')
        #
        res_g.compute_traj()
        # imtio.export_to_file(res_g, "VF1_compute_traj1_g.cimt")
        res_g2 = imtio.import_from_file("VF1_compute_traj1_g.cimt")
        assert res_g == res_g2
        #
        res_g.compute_traj(epsilon=4)
        # imtio.export_to_file(res_g, "VF1_compute_traj2_g.cimt")
        res_g2 = imtio.import_from_file("VF1_compute_traj2_g.cimt")
        assert res_g == res_g2
        #
        res_g.compute_traj(epsilon=4*make_unit('m'))
        # imtio.export_to_file(res_g, "VF1_compute_traj3_g.cimt")
        res_g2 = imtio.import_from_file("VF1_compute_traj3_g.cimt")
        assert res_g == res_g2
        #
        res_g.compute_traj(close_traj=True)
        # imtio.export_to_file(res_g, "VF1_compute_traj4_g.cimt")
        res_g2 = imtio.import_from_file("VF1_compute_traj4_g.cimt")
        assert res_g == res_g2

    def test_get_point_density(self):
        cps = vod.get_critical_points(self.TVF1_nomask, kind='pbi_crit',
                                      thread='all')
        #
        dens = cps.get_points_density('foc')
        # imtio.export_to_file(dens, "VF1_get_points_density_a.cimt")
        dens_2 = imtio.import_from_file("VF1_get_points_density_a.cimt")
        assert dens == dens_2

    def test_break_trajectories(self):
        cps = vod.get_critical_points(self.TVF1_nomask, kind='pbi_crit',
                                      thread='all')
        cps.compute_traj()
        cps.break_trajectories()
        # imtio.export_to_file(cps, "VF1_break_trajectories_a.cimt")
        cps_2 = imtio.import_from_file("VF1_break_trajectories_a.cimt")
        assert cps == cps_2

    def test_get_mean_trajectory(self):
        cps = vod.get_critical_points(self.TVF1_nomask, kind='pbi_crit',
                                      thread='all')
        cps.compute_traj()
        means = cps.get_mean_trajectory("foc", min_len=0, min_nmb_to_avg=0)
        # imtio.export_to_file(means, "VF1_get_mean_trajectory_a.cimt")
        means_2 = imtio.import_from_file("VF1_get_mean_trajectory_a.cimt")
        assert means == means_2

    def test_get_vortex_position(self):
        fun = vod.get_vortex_position
        kwargs = [{'obj': self.VF1_nomask,
                   'criterion': voc.get_residual_vorticity},
                  {'obj': self.VF1_nomask,
                   'criterion': voc.get_residual_vorticity,
                   'threshold': 0.2}]
        parametric_test(fun, kwargs)

    def test_mean_trajectories(self):
        pass

    def test_crit_points(self):
        pass

    def test_topo_points(self):
        pass

# TEMP
pytest.main(['test_vortex_detection.py'])
# pytest.main(['--pdb', 'test_vortex_detection.py'])
# test = TestVortexDetection()
# test.setup()
# test.test_get_vortex_position()
# TEMP - End



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from multiprocess import Pool, cpu_count, Manager
from . import ProgressCounter


class MultiThreading(object):
    def __init__(self, funct, data, threads='all'):
        raise Exception("Not functionnal yet !")
        self.funct = funct
        if threads == 'all':
            threads = cpu_count()
        self.pool = Pool(processes=threads)
        self.data = data
        self.PG = None
        self.initializer = None
        self.finalizer = None

    def add_progress_counter(self, init_mess="Beginning", end_mess="Done",
                             name_things='things', perc_interv=5):
        self.PG = ProgressCounter(init_mess=init_mess, end_mess=end_mess,
                                  nmb_max=len(self.data),
                                  name_things=name_things,
                                  perc_interv=perc_interv)
        self.manager = Manager()
        self.manager.register("PG", self.PG)

    def run(self):
        res = self.pool.map_async(self.PG_func_wrapper, self.data)
        self.pool.close()
        self.pool.join()
        return res
